/*-
 * Copyright (c) 1999, 2000, 2001 Lev Walkin <vlm@spelio.net.ru>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: cfgread.c,v 1.3 2001/05/15 15:19:23 vlm Exp $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef	HAVE_MMAP
#include <sys/mman.h>
#endif

#include "strfunc.h"

char *sf_cfg_i_cf;
char *sf_cfg_i_cfpos;
off_t	sf_cfg_i_cfsize;

int sfcfgparse();

extern int sf_cfg_i_nline;
extern int sf_cfg_i_brcount;

int cfgread(const char *cfname) {
	int fd;
	struct stat sb;
#ifndef	HAVE_MMAP
	ssize_t rd, rds;
	char *buf;
#endif

	if(!cfname) {
		fprintf(stderr, "-ERR: filename required.\n");
		return -1;
	};

	if((fd=open(cfname, O_RDONLY)) == -1 || fstat(fd, &sb)) {
		fprintf(stderr, "-ERR: Can't open config file.\n");
		return -1;
	};

	sf_cfg_i_cfsize=sb.st_size+1;
	
#ifdef	HAVE_MMAP
	if(!(sf_cfg_i_cfpos=sf_cfg_i_cf=(char *)mmap(NULL, sf_cfg_i_cfsize, PROT_READ | PROT_WRITE, MAP_PRIVATE, fd, 0))) {
		close(fd);
		fprintf(stderr, "-ERR: Can't mmap() config file.\n");
		return -1;
	}
#else
	buf = sf_malloc(sb.st_size + 1);
	rds = 0;
	while(rds < sb.st_size) {

		rd = read(fd, buf + rds, sb.st_size - rds);
		if(rd == -1) {
			free(buf);
			fprintf(stderr, "-ERR: Can't correctly read() config file.\n");
			return -1;
		};
		rds += rd;
	};

	buf[rds] = 0;

	sf_cfg_i_cfpos=sf_cfg_i_cf=buf;
#endif

	sf_cfg_i_nline=0;
	sf_cfg_i_brcount=0;

	if(sfcfgparse()) {
		close(fd);
#ifdef	HAVE_MMAP
		munmap(sf_cfg_i_cf, sf_cfg_i_cfsize);
#else
		free(sf_cfg_i_cf);
#endif
		return 1;
	};

	close(fd);
#ifdef	HAVE_MMAP
	munmap(sf_cfg_i_cf, sf_cfg_i_cfsize);
#else
	free(sf_cfg_i_cf);
#endif

	return 0;
};

extern slist *sf_cfg_i_a, *sf_cfg_i_v;

char *cfgget(char *at) {
	return sget2(sf_cfg_i_a, at, sf_cfg_i_v);
};

slist *cfgget2(char *at) {
	slist *sl=NULL;
	int i;

	if(!sf_cfg_i_a) return NULL;

	sl = sinit();

	if(!at)
		return sl;

	for(i=0; i< sf_cfg_i_a->count; i++)
		if(ecq(sf_cfg_i_a->list[i], at))
			sadd(sl, sf_cfg_i_v->list[i]);

	return sl;
};

