/*-
 * Copyright (c) 1999, 2000, 2001 Lev Walkin <vlm@spelio.net.ru>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: cgifunc.c,v 1.13 2001/05/15 15:19:23 vlm Exp $
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include <errno.h>

#include "strfunc.h"

char *_sf_urld_buf=NULL;
ssize_t _sf_urld_buflen=0;

extern int _sf_add_internal(slist *, void *msg, size_t len);

char *
url_decode(char *str) {
    char *buf;
	char *p;
    int n;
	unsigned char c;

	n = strlen(str?str:"") + 1;
	if(n < _sf_urld_buflen)
		n = _sf_urld_buflen;

	p = buf = (char *)sf_malloc(n);

	if(str) {
	  for(; (c=*str); str++, p++) {
        if(c == '%') {
            *p = 0;
			c = *++str; 
            if(!c || !*(str+1)) break;

			if( c >= '0' && c <= '9') *p=c-'0';
			else if( (c >= 'A' && c <= 'F') || (c >= 'a' && c <= 'f') )
				*p=10+c-((c>='a')?'a':'A');

            *p*=16;
			c=*++str;

			if( c >= '0' && c <= '9') *p+=c-'0';
			else if( (c >= 'A' && c <= 'F') || (c >= 'a' && c <= 'f') )
				*p+=10+c-((c>='a')?'a':'A');

        } else {
			*p=(c=='+')?' ':c;
		}
      } /* while */
	} /* if */
    *p=0;

	if(_sf_urld_buf)
		free(_sf_urld_buf);
	_sf_urld_buflen = n;
    return (_sf_urld_buf = buf);
};

char *_sf_urle_buf=NULL;
int _sf_urle_buflen=0;

char *
url_encode(char *str) {
	char *buf;
	char *p;
	static char *table="0123456789ABCDEF";
	int n;
	unsigned char c;

	n = strlen(str?str:"") + 1;
	if(_sf_urle_buflen > n)
		n = _sf_urle_buflen;

	p = buf = (char *)sf_malloc((n + 2) * 3);

	if(str) {
		for(; (c=*str); str++) {
			if(
				(c >= 'A' && c <= 'Z') ||
				(c >= 'a' && c <= 'z') ||
				(c >= '0' && c <= '9') ||
				strchr("@._", *str)
			) {
				*p++ = *str;
				continue;
			}
			if(c == ' ') {
				*p++ = '+';
				continue;
			};
			*p++ = '%';
			*p++ = table[ (unsigned char)(c & 0xF0) >> 4];
			*p++ = table[ (unsigned char)(c & 0x0F) ];
		} /* for */
	} /* if(str) */
	*p=0;

	if(_sf_urle_buf)
		free(_sf_urle_buf);
	_sf_urle_buflen = n;
    return (_sf_urle_buf = buf);
};

slist *_sf_attr=NULL;
slist *_sf_vals=NULL;
slist *_sf_unmv=NULL; /* Unmodified value */
slist *_sf_type=NULL; /* Content-type */

int _sf_cgi_parse_multipart(char *data, size_t len);

int
parse_form() {
	char *s;
	slist *sl;
	int n;
	char *af, *as;
	static char parsed=0;

	if(parsed) return 0;

	if(!_sf_attr) _sf_attr=sinit();
	if(!_sf_vals) _sf_vals=sinit();
	if(!_sf_unmv) _sf_unmv=sinit();
	if(!_sf_type) _sf_type=sinit();

	sl = sinit();
	if(!sl)
		return -1;

	if(!(s=getenv("REQUEST_METHOD")) ||
		(strcmp(s, "GET") && strcmp(s, "HEAD") && strcmp(s, "POST")))
			return -1;

	if(strcmp(s, "POST")) {
		if(!(s=getenv("QUERY_STRING")))
			return -1;
		splitf(sl, s, "&", 0);
	} else {
		char *buf;
		ssize_t buflen, bl, tmp;

		if(!(s=getenv("CONTENT_LENGTH")) || (buflen=atoi(s)) < 0)
			buflen=4092; // probably, one page.

		buf=(char *)sf_malloc(buflen+4);

		bl=0;
		while(bl < buflen) {
			if((tmp=read(0, buf+bl, buflen - bl)) == -1) {
				if(errno == EINTR)
					continue;
				free(buf);
				return -1;
			}
			if(!tmp)
				return -1;
			bl+=tmp;
		};

		buf[buflen=bl]=0;

		/* Parse differently, if it is multipart form. */
		if((s=getenv("CONTENT_TYPE")) &&
			!strncasecmp(s, "multipart/form-data",
				sizeof("multipart/form-data")-1)) {
			buflen=_sf_cgi_parse_multipart(buf, buflen);
			free(buf);
			return 0;
		};

		splitf(sl, buf, "&", 0);
		free(buf);
	};


	if(!sl->count) {
		sfree(sl);
		return 0;
	};

	for(n=0; n<sl->count; n++) {
		af=sl->list[n];
		if((as=strchr(af, '='))) *as++=0;
		sadd(_sf_attr, url_decode(af));
		sadd(_sf_unmv, as?as:"");	/* Unmodified value */
		sadd(_sf_vals, url_decode(as));	/* Decoded value */
		sadd(_sf_type, "text/unknown");
	};

	sfree(sl);

	parsed = 1;
	return 0;
};

char *
param(char *field) {
	if(!_sf_attr)
		parse_form();
	return scget2(_sf_attr, field, _sf_vals);
};

slist *
param2(char *field, int flag) {
	static slist *sl=NULL;
	slist *vals;
	int n;

	if(!_sf_attr)
		parse_form();

	if(!sl) {
		sl=sinit();
		if(!sl)
			return NULL;
	} else {
		sl->count = 0;
		sl->maxlen = 0;
		sl->list[0] = NULL;
		sl->lens[0] = 0;
	}

	if(!field)
		return sl;

	switch(flag) {
		case 0:
			vals = _sf_vals;
			break;
		case 1:
			vals = _sf_unmv;
			break;
		case 2:
			vals = _sf_type;
			break;
		default:
			vals = _sf_vals;
	}

	for(n=0; n < _sf_attr->count; n++)
		if(ecq(_sf_attr->list[n], field))
			_sf_add_internal(sl, vals->list[n], vals->lens[n]);

	return sl;
};

slist *
params() {
	return _sf_attr;
};

static char *_sf_hq_buf=NULL;
static int _sf_hq_len=0;

char *
html_quote_detach() {
	char *buf = _sf_hq_buf;
	_sf_hq_buf=NULL;
	return buf;
};

size_t
html_quote_lastsize() {
	return _sf_hq_len;
};

char *
html_quote(char *s) {
	char *p;
	int n;

	if(!(n=countchar2(s, "\"&<>"))) return s;
	n=strlen(s) + sizeof("&quot;") * n + 1;
	if(n >= _sf_hq_len)
		_sf_hq_buf=(char *)sf_realloc(_sf_hq_buf, _sf_hq_len=n);
	p=_sf_hq_buf;

	while(1)
	switch(*s) {
		case 0:
			*p=0; return _sf_hq_buf;
		case '"':
			strcpy(p, "&quot;"); p+=6; s++; break;
		case '<':
			strcpy(p, "&lt;"); p+=4; s++; break;
		case '>':
			strcpy(p, "&gt;"); p+=4; s++; break;
		case '&':
			strcpy(p, "&amp;"); p+=5; s++; break;
		default:
			*p++ = *s++; continue;
	};
};

char *_sfqpbuf=NULL;
int _sf_in_mime_words;

char *_sf_b64_buf=NULL;
size_t _sf_b64_len=0;

void *
base64_detach(size_t *size) {
	void *buf=_sf_b64_buf;
	_sf_b64_buf=NULL;
	if(size)
		*size = _sf_b64_len;
	return (void *)buf;
};

char *
quoted_printable_decode(char *p) {
	char *buf;
	char *s;
	register int n;
    if(!p) {
		if(_sfqpbuf)
			free(_sfqpbuf);
		_sfqpbuf=sf_strdup("");
		return _sfqpbuf;
	};
	s=buf=(char *)sf_malloc(strlen(p) + 1);

	p--;
	while(*(++p)) {
		if(_sf_in_mime_words && (*p == '_')) {
			*s++ = ' ';
			continue;
		};
        if(*p == '=') {
            *s=0;
            n=*++p; 
            if(!n) { p--; break; }
			if(!*(p+1)) break;

			if( n >= '0' && n <= '9') *s=n-'0';
			else if( (n >= 'A' && n <= 'F') || (n >= 'a' && n <= 'f') )
				*s=10+n-((n>='a')?'a':'A');
			else {
				if(n == '\n' || n == '\r') continue;
				*s++ = '='; *s++ = n; continue;
			};

			*s*=16;
			n=*++p;

			if( n >= '0' && n <= '9') *s += n-'0';
			else if( (n >= 'A' && n <= 'F') || (n >= 'a' && n <= 'f') )
				*s += 10+n-((n>='a')?'a':'A');
			else {
				if(n == '\n' || n == '\r') { *s++ = n; continue; };
				*s = ' '; continue;
			};

			s++;
			continue;
        };
		*s++=*p;
    }
    *s=0;
	if(_sfqpbuf) free(_sfqpbuf);
    return _sfqpbuf=buf;
};


char *
_sf_handler(slist *sl) {
	char *p;
	if(sl->count < 4) {
		sclear(sl);
		return NULL;
	}

	p = sl->list[3];
	if(ecq(sl->list[2], "Q")) {
		_sf_in_mime_words++;
		p = quoted_printable_decode(p);
		_sf_in_mime_words--;
	} else if(ecq(sl->list[2], "B")) {
		p = (char *)base64_decode(p, NULL);
	};

	if(ecq(sl->list[1], "windows-1251")) {
		p = windows_1251_to_koi8_r(p);
	} else if(ecq(sl->list[1], "iso-8859-5")) {
		p = iso_8859_5_to_koi8_r(p);
	} else if(ecq(sl->list[1], "cp-866")) {
		p = cp_866_to_koi8_r(p);
	};

	sclear(sl);

	return p;
};

char *
mime_word_decode(char *str) {
	static sed_t *se;
	static sbuf *sbs;
	sbuf *sb;

	if(!se) {
		se = sed_compile("s/=\\?([a-z0-9._-]+)\\?(.)\\?([^[:space:]\n\r\t?]+)\\?=[ \n\r\t]*/\\@/gei");
		if(!se)
			fprintf(stderr, "STRFUNC: Can't compile expression in mime_word_decode().\n");
		sed_handler(se, _sf_handler);
	};

	if(!sbs) {
		sbs = sbuf_init();
		if(!sbs)
			return NULL;
	}

	/* Dynamic allocation */

	sb = sbuf_init();
	if(!sb)
		return NULL;

	if(!str) {
		if(sbuf_add(sb, "") == -1) {
			sbuf_free(sb);
			return NULL;
		}
		goto mwde;
	};

	if(countchar2(str, "=?") < 4) {
		sbuf_add(sb, str);
		goto mwde;
	};


	if(sbuf_add(sb, sed_exec(se, str)) == -1) {
		sbuf_free(sb);
		return NULL;
	};

mwde:
	if(sbs)
		sbuf_free(sbs);
	sbs = sb;
	return sbs->buf;
};

static unsigned char *_sf_ucp_bi=NULL;
static unsigned char _sf_uc_ib[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/==";

void
_sf_b64_init() {
	if(!_sf_ucp_bi) {
		_sf_ucp_bi=(char *)sf_malloc(256);
	};
	memset(_sf_ucp_bi, 100, 256);
	_sf_ucp_bi['A']=0;
	_sf_ucp_bi['B']=1;
	_sf_ucp_bi['C']=2;
	_sf_ucp_bi['D']=3;
	_sf_ucp_bi['E']=4;
	_sf_ucp_bi['F']=5;
	_sf_ucp_bi['G']=6;
	_sf_ucp_bi['H']=7;
	_sf_ucp_bi['I']=8;
	_sf_ucp_bi['J']=9;
	_sf_ucp_bi['K']=10;
	_sf_ucp_bi['L']=11;
	_sf_ucp_bi['M']=12;
	_sf_ucp_bi['N']=13;
	_sf_ucp_bi['O']=14;
	_sf_ucp_bi['P']=15;
	_sf_ucp_bi['Q']=16;
	_sf_ucp_bi['R']=17;
	_sf_ucp_bi['S']=18;
	_sf_ucp_bi['T']=19;
	_sf_ucp_bi['U']=20;
	_sf_ucp_bi['V']=21;
	_sf_ucp_bi['W']=22;
	_sf_ucp_bi['X']=23;
	_sf_ucp_bi['Y']=24;
	_sf_ucp_bi['Z']=25;

	_sf_ucp_bi['a']=26;
	_sf_ucp_bi['b']=27;
	_sf_ucp_bi['c']=28;
	_sf_ucp_bi['d']=29;
	_sf_ucp_bi['e']=30;
	_sf_ucp_bi['f']=31;
	_sf_ucp_bi['g']=32;
	_sf_ucp_bi['h']=33;
	_sf_ucp_bi['i']=34;
	_sf_ucp_bi['j']=35;
	_sf_ucp_bi['k']=36;
	_sf_ucp_bi['l']=37;
	_sf_ucp_bi['m']=38;
	_sf_ucp_bi['n']=39;
	_sf_ucp_bi['o']=40;
	_sf_ucp_bi['p']=41;
	_sf_ucp_bi['q']=42;
	_sf_ucp_bi['r']=43;
	_sf_ucp_bi['s']=44;
	_sf_ucp_bi['t']=45;
	_sf_ucp_bi['u']=46;
	_sf_ucp_bi['v']=47;
	_sf_ucp_bi['w']=48;
	_sf_ucp_bi['x']=49;
	_sf_ucp_bi['y']=50;
	_sf_ucp_bi['z']=51;

	_sf_ucp_bi['0']=52;
	_sf_ucp_bi['1']=53;
	_sf_ucp_bi['2']=54;
	_sf_ucp_bi['3']=55;
	_sf_ucp_bi['4']=56;
	_sf_ucp_bi['5']=57;
	_sf_ucp_bi['6']=58;
	_sf_ucp_bi['7']=59;
	_sf_ucp_bi['8']=60;
	_sf_ucp_bi['9']=61;
	_sf_ucp_bi['+']=62;
	_sf_ucp_bi['/']=63;

	_sf_ucp_bi[0]=255;
	_sf_ucp_bi['=']=255;
};

char *base64_encode(void *data, size_t size) {
	char *output;
	register char *ou;
	register unsigned char *p=(unsigned char *)data;
	register void *dte=data + size;
	register int nc=0;

	if(!data) return NULL;

	ou=output=(char *)sf_malloc(size / 3 * 4 + (size / 50) + 1);

	while((char *)dte - (char *)p >= 3) {
		*ou = _sf_uc_ib[ *p >> 2 ];
		ou[1] = _sf_uc_ib[ ((*p & 0x03) << 4) | (p[1] >> 4) ];
		ou[2] = _sf_uc_ib[ ((p[1] & 0x0F) << 2) | (p[2] >> 6) ];
		ou[3] = _sf_uc_ib[ p[2] & 0x3F ];

		p+=3;
		ou+=4;

		nc+=4;
		if(!(nc % 76)) *ou++='\n';
	};
	if((char *)dte - (char *)p == 2) {
		*ou++ = _sf_uc_ib[ *p >> 2 ];
		*ou++ = _sf_uc_ib[ ((*p & 0x03) << 4) | (p[1] >> 4) ];
		*ou++ = _sf_uc_ib[ ((p[1] & 0x0F) << 2) ];
		*ou++ = '=';
	} else if((char *)dte - (char *)p == 1) {
		*ou++ = _sf_uc_ib[ *p >> 2 ];
		*ou++ = _sf_uc_ib[ ((*p & 0x03) << 4) ];
		*ou++ = '=';
		*ou++ = '=';
	};

	*ou=0;

	_sf_b64_len = (ou - output);

	if(_sf_b64_buf)
		free(_sf_b64_buf);
	return _sf_b64_buf=output;
};

void *
base64_decode(char *str, size_t *size) {
	static int init=0;
	void *output, *ou;

	if(size) *size=0;
	if(!str) return NULL;

	if(!init) { _sf_b64_init(); init=1; };

	output=ou=(void *)sf_malloc(strlen(str) + 1);

{
	unsigned int pr[6];
	register int n;
	register int doit=1;

	while(doit) {
		n=0;
		while(n < 4) {
			register unsigned char ch;
			ch=_sf_ucp_bi[*(unsigned char *)str];
			if(ch < 100) { pr[n]=ch; n++; str++; continue; };
			if(ch == 100) { str++; continue; };
			doit=0; break;
		};
		if(!doit && n < 4) { pr[n+2]=pr[n+1]=pr[n]=0; };

		*(char *)ou=(pr[0] << 2) | (pr[1] >> 4);
		*(char *)(ou+1)=(pr[1] << 4) | (pr[2] >> 2);
		*(char *)(ou+2)=(pr[2] << 6) | pr[3];

		ou+=(n * 3) >> 2 ;
		
	};

	*(char *)ou=0;
}

	_sf_b64_len = (ou - output);

	if(size) *size=_sf_b64_len;
	if(_sf_b64_buf)
		free(_sf_b64_buf);
	return (_sf_b64_buf=output);
};

int
_sf_cgi_parse_multipart(char *data, size_t len) {
	char *bnd;
	int bndlen;
	char *dp=data;
	char *sd=NULL;	/* start data */
	char *fname=NULL;
	char *name=NULL;
	char *ct=NULL;
	char *tmp;
	slist *sl1, *sl2;
	int i, z;

	if(!dp)
		return 0;
	if(!(bnd=getenv("CONTENT_TYPE")))
		return 0;
	if(!(bnd=strstr(bnd, "boundary=")))
		return 0;
	bnd+=sizeof("boundary")-2;

	bnd=sf_strdup(bnd);
	if(!bnd) return 0;
	*bnd=bnd[1]='-';
	bndlen=strlen(bnd);

	sl1=sinit();
	sl2=sinit();

	while(dp-data < len) {
		if(strncmp(dp, bnd, bndlen)) { dp++; continue; };

		*(dp-1)=*(dp-2)=0;

		if(sd) {
			sadd(_sf_attr, name?name:"UNKNOWN");
			sadd(_sf_type, ct?ct:"");
			if(fname) {
				sadd(_sf_vals, fname);
				sadd2(_sf_unmv, sd, dp-sd-2);
			} else {
				sadd2(_sf_vals, sd, dp-sd-2);
				sadd2(_sf_unmv, sd, dp-sd-2);
			};
			if(fname) free(fname);
			if(name) free(name);
			fname=name=sd=ct=NULL;
		};

		dp+=bndlen;
		if(!strncmp(dp, "--\r\n", 4)) break;
		dp+=2;

		/* Part started */

		tmp=strstr(dp, "\r\n\r\n");
		tmp+=2;
		*tmp=0;
		sd=tmp+2;

		tmp=dp-1;
		while(*(++tmp)) if(*tmp == ';') *tmp=' ';

		sclear(sl1);
		splitf(sl1, dp, "\r\n", 0);
		if(!sl1->count) continue;

		for(i=0; i<sl1->count; i++) {
			if(!strncasecmp(sl1->list[i], "Content-Disposition:",
				sizeof("Content-Disposition:")-1)) {
				sclear(sl2);
				splitquotable(sl2, sl1->list[i]);
				for(z=0; z<sl2->count; z++) {
					if(!strncasecmp(sl2->list[z], "name=", 5))
						name=sl2->list[z]+5;
					else
					if(!strncasecmp(sl2->list[z], "filename=", 9))
						fname=sl2->list[z]+9;
				};
			} else if(!strncasecmp(sl1->list[i], "Content-Type:",
				sizeof("Content-Type:")-1)) {
				ct=sl1->list[i] + sizeof("Content-Type:") - 1;
				while(*ct == ' ') ct++;
			};
		};

	};

	sfree(sl1);
	sfree(sl2);

	return 1;
};

int
set_cookie(char *name, char *value, char *optDomain, char *optPath, time_t optMaxAge, char *optComment, int optSecure) {
	if(!name || !value || !*name) return -1;

	printf("Set-Cookie: %s=\"%s\"", name, url_encode(value));
	if(optDomain) printf("; Domain=%s", optDomain);
	if(optPath) printf("; path=\"%s\"", optPath);
	if(optMaxAge >=0 ) printf("; Max-Age=%ld", optMaxAge);
	if(optComment) printf("; Comment=\"%s\"", html_quote(optComment));
	if(optSecure) printf("; Secure");
	printf("\n");
	return 0;
};

slist *sf_cgi_csla=NULL;
slist *sf_cgi_cslv=NULL;

slist *cookies() {
	if(!sf_cgi_csla) cookie("");
	return sf_cgi_csla;
};

char *cookie(char *name) {
	register char *p, *s;

	if(!name) return NULL;

	if(!sf_cgi_csla) {
		char *pt=getenv("HTTP_COOKIE");
		if(!pt || !*pt) return NULL;
		s=p=pt=sf_strdup(pt);

		sf_cgi_csla=sinit();
		sf_cgi_cslv=sinit();

		p--;
		while(*++p) {
			if(*p == '=') {
				*p++='\0';
				sadd(sf_cgi_csla, s);
				s=p;
				if(*s == '"') p=strchr(++s, '"');
				else p=strchr(s, ';');
				if(!p) p=strrchr(s, '\0');
				*p++=0;
				sadd(sf_cgi_cslv, url_decode(s));
				if(*p == '"') p++;
				if(*p == ';') p++;
				while(*p == ' ') p++;
				s=p;
			};
		};
		free(pt);
	};

	if(!sf_cgi_csla->count)
		return NULL;

	return scget2(sf_cgi_csla, name, sf_cgi_cslv);

	return NULL;
};

char *
limittextwidth(char *txt, size_t width) {
	register int n;
	register char *c, *tc;

	if(!txt) return NULL;
	if(!*txt) return txt;
	if((ssize_t)width < 1) return txt;

	n=0;
	c=txt;

	while(*++c) {
		if(*c == '\n') { n=0; continue; };
		if(++n <= width) continue;

		tc=c;
		while((c > txt) && (*c != '\n')) {
			if(*--c == ' ') {
				*c='\n';
				n=0;
				tc=NULL;
				break;
			};
		};

		if(!tc) {
			n=0;
			continue;
		};

		while(*++c && *c != ' ');
		n=0;
		if(*c) *c='\n';
		else return txt;
	};

	return txt;
};


slist *
getlanguageprefs() {
	static slist *sl=NULL;
	char *p;
	int i;
	register char *t;

	if(sl) return sl;

	if(!(p=getenv("HTTP_ACCEPT_LANGUAGE")))
		return NULL;

	sl = split(p, ", ", 0);

	for(i=0; i<sl->count; i++) {
		if((p=strchr(sl->list[i], ';'))) {
			*p=0;
			if((p - sl->list[i]) == 0) {
				sdel(sl, i--);
				continue;
			};
		};

		t = sl->list[i];
		for(t=sl->list[i]; *t; t++) {
			if(*t == '-')
				continue;
			if(*t >= 'a' && *t <= 'z')
				continue;
			if(*t >= '0' && *t <= '9')
				continue;
			if(*t >= 'A' && *t <= 'Z')
				continue;
			break;
		};
		if(*t)	/* Not reached the end */
			sdel(sl, i--);
	};

	if(!sl->count) { sfree(sl); sl=NULL; return NULL; };

	return sl;
};

