/*-
 * Copyright (c) 1999, 2000, 2001 Lev Walkin <vlm@spelio.net.ru>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: datetime.c,v 1.7 2001/05/15 15:19:23 vlm Exp $
 */

#ifndef	__EXTENSIONS__
#define	__EXTENSIONS__
#endif

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/types.h>
#include <time.h>
#include <regex.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include "strfunc.h"

#ifndef	HAVE_TM_ZONE
#ifndef	HAVE_TZNAME
#error	"Please, please report this situation to vlm@spelio.net.ru!"
#error	"This iz due to my lazyness in checking the unusual states..."
#endif
#endif

int
_sf_isspace(char c) {
	if(c == ' ' || c == '\t' || c == '\n' || c == '\r')
		return 1;
	return 0;
};

/* Water-proof atoi() */
inline int
_sf_natoui(char *str, int num) {
	register int n=0, na;
	register char *s = str;

	if(!str) return 0;

	while(num-- > 0) {
		na = *s++ - '0';
		if(na < 0 || na > 9) return -1;
		n = (n * 10) + na;
	};

	return n;
};

inline int
_sf_recognize_month(char *dts) {
	switch(*dts) {
		case 'F': return 1;
		case 'M':
			if(!strncmp(dts, "Mar", 3))
				return 2;
			else if(!strncmp(dts, "May", 3))
				return 4;
			break;
		case 'A':
			if(!strncmp(dts, "Apr", 3)) return 3;
			else if(!strncmp(dts, "Aug", 3)) return 7;
			break;
		case 'J':
			if(!strncmp(dts, "Jan", 3)) return 0;
			else if(!strncmp(dts, "Jun", 3)) return 5; 
			else if(!strncmp(dts, "Jul", 3)) return 6;
			break;
		case 'S': return 8;
		case 'O': return 9;
		case 'N': return 10;
		case 'D': return 11;
	};

	return -1;
};

inline time_t
_sf_recognize_tz(char *dts) {
	time_t tmp=0;

	if(*dts == '+' || *dts == '-') {
		if((tmp = _sf_natoui(dts+1, 4)) < 0) return -1;
		tmp = (tmp / 100) * 3600 + (tmp % 100) * 60;
		if(*dts == '-')
			tmp = -tmp;
		dts += 5;
		while(*dts == ' ') dts++;
		return tmp;
	} else if(*dts == '(' || (*dts >= 'A' && *dts <= 'Z')) {

		if(*dts == '(') dts++;

		if(!*dts || !dts[1] || !dts[2]
		|| (dts[3] >= 'A' && dts[3] <= 'Z'))
			return 0;

		/* RFC 822 zones */
		if(!strncmp(dts, "GMT", 3))
			tmp = 0;
		else if(!strncmp(dts, "EST", 3))
			tmp = - 5 * 3600;
		else if(!strncmp(dts, "EDT", 3))
			tmp = - 4 * 3600;
		else if(!strncmp(dts, "CST", 3))
			tmp = - 6 * 3600;
		else if(!strncmp(dts, "CDT", 3))
			tmp = - 5 * 3600;
		else if(!strncmp(dts, "MST", 3))
			tmp = - 7 * 3600;
		else if(!strncmp(dts, "MDT", 3))
			tmp = - 6 * 3600;
		else if(!strncmp(dts, "PST", 3))
			tmp = - 8 * 3600;
		else if(!strncmp(dts, "PDT", 3))
			tmp = - 7 * 3600;
		else return -1;

		return tmp;
	};

	return -1;
};


inline long
_sf_recognize_time(char *dts) {
	long l = 0;
	int tmp;

	if(!*dts)
		return -1;

	if((tmp = _sf_natoui(dts, 2)) < 0 || tmp > 24)
		return -1;

	l |= tmp << 16;
	if(dts[2] != ':' && dts[2] != '-') return -1;
	dts+=3;

	if((tmp = _sf_natoui(dts, 2)) < 0 || tmp > 60)
		return -1;

	l |= tmp << 8;
	if(dts[2] != ':' && dts[2] != '-') {
		l |= 5 << 24;
		return -1;
	}
	dts+=3;

	if((tmp = _sf_natoui(dts, 2)) < 0 || tmp > 60)
		return -1;

	l |= tmp;
	l |= 8 << 24;

	return l;
}

time_t
strtotime(char *dts) {
	struct tm tm;
	time_t tmp=0;
	int form;

	if(!dts) return 0;

	/* That's a good start. */
	while(*dts == ' ')
		dts++;

	/*
	** Forms:
	** form = 0: something like error
	** form = 1: ? RFC 822
	** form = 2: ? ISO 8601
	** form = 3: ? ctime()
	** form = 4: ? syslogd format (Mar 23 18:15:00)
	*/

	form = 0;
	if(*dts >= 'A') {
		form = 1;
	} else if((*dts >= '0' && *dts <= '9')) {
		int n = 0;
		while(dts[n] >= '0' && dts[n] <= '9') n++;
		if(!dts[n] || (_sf_isspace(dts[n]) && !dts[n+1])) {
			if(atoi(dts) < 0x7fffffff)
				return (time_t)atol(dts);
		};
		form = 2;
	} else
		return 0;


	/* Kepka Monomaha */
	memset(&tm, 0, sizeof(struct tm));
	tm.tm_isdst=-1;
	tm.tm_mday=1;


if(form == 1) {
	char *savedts=dts;

	while(*dts && *dts != ',' && *dts != ' ')
		dts++;

	if(*dts != ',') {
		/* Maybe, it is just ctime()? */
		dts = savedts;
		form = 3;
		goto f3;
	};

	dts++;
	while(*dts == ' ') dts++;
	if(!*dts) return 0;

	if(*dts && dts[1] == ' ') {
		tm.tm_mday = _sf_natoui(dts++, 1);
	} else {
		tm.tm_mday = _sf_natoui(dts, 2);
		dts+=2;
	};

	if(tm.tm_mday < 1)
		return 0;

	while(*dts == ' ' || *dts == '-') dts++;
	if(!*dts) return 0;

	tm.tm_mon = _sf_recognize_month(dts);

	while(*dts >= 'A' || *dts == ' ' || *dts == '-') dts++;

	if(!*dts) return 0;

	if((tm.tm_year = _sf_natoui(dts, 4)) < 0) {
		if((tm.tm_year = _sf_natoui(dts, 2)) < 0)
			return 0;
		if(tm.tm_year < 38)
			tm.tm_year += 100;
		dts+=2;
	} else {
		tm.tm_year -= 1900;
		if(tm.tm_year < 0)
			return 0;
		dts+=4;
	};

	while(*dts == ' ') dts++;

	if((tmp = _sf_recognize_time(dts)) == -1)
		return 0;

	dts += ((unsigned int)tmp) >> 24;
	tm.tm_hour = (tmp >> 16) & 0xff;
	tm.tm_min = (tmp >> 8) & 0xff;
	tm.tm_sec = tmp & 0xff;
	
	while(*dts == ' ') dts++;

	tmp = _sf_recognize_tz(dts);

	goto ex;

};

if(form == 2) {

	/* Year */
	if((tm.tm_year = _sf_natoui(dts, 4)) == -1) {
		/* hh:mm... form */
		if(strlen(dts) >= 5) {
			if(dts[2] == ':') {
				time_t tloc;
				time(&tloc);
				localtime_r(&tloc, &tm);
				tm.tm_isdst = -1;
				tm.tm_hour = 0; tm.tm_sec = 0; tm.tm_min = 0;
				goto tme;
			} else if(dts[2] == '/' || dts[2] == '-') {
				time_t tloc;
				time(&tloc);
				localtime_r(&tloc, &tm);
				tm.tm_isdst = -1;
				tm.tm_mon = 0; tm.tm_mday = 0;
				tm.tm_hour = 0; tm.tm_sec = 0; tm.tm_min = 0;
				goto dte;
			};
		};
		return 0;
	};

	tm.tm_year -= 1900;

	dts+=4;
	if(!*dts) return 0;

	if(*dts == '-' || *dts == '/') dts++;
	else {
		if(*dts == '+' || *dts == 'Z') goto zone;
		if(*dts == 'T') {
			dts++;
			goto tme;
		};
	};

dte:

	/* Month */
	if((tm.tm_mon = _sf_natoui(dts, 2)) == -1) {
		if(dts[-1] == '-') return 0;
		if(!_sf_isspace(*dts)) return 0;
		goto ex;
	};

	tm.tm_mon--;

	if(tm.tm_mon < 0)
		return 0;

	dts+=2;
	if(!*dts) goto ex;

	if(*dts == '-' || *dts == '/') dts++;
	else {
		if(*dts == '+' || *dts == 'Z') goto zone;
		if(*dts == 'T') {
			dts++;
			goto tme;
		};
	};

	/* Month Day */
	if((tm.tm_mday = _sf_natoui(dts, 2)) == -1) {
		if(dts[-1] == '-') return 0;
		if(!_sf_isspace(*dts)) return 0;
		goto ex;
	};

	if(!tm.tm_mday)
		return 0;

	dts+=2;

lp:
	if(!*dts) goto ex;
	if(*dts == 'T') {
		dts++;
		goto tme;
	};
	if(*dts >= '0' && *dts <= '9')
		goto tme;
	if(*dts == '+') goto zone;
	if(*dts == 'Z') goto ex;
	if(*dts++ == ' ')
		goto lp;
	goto ex;

tme:

	if((tm.tm_hour = _sf_natoui(dts, 2)) == -1)
		return 0;

	dts+=2;

	if(!*dts) goto ex;
	if(*dts == ':') dts++;
	else {
		if(*dts == '+' || *dts == '-') goto zone;
		if(*dts == 'Z') goto ex;
	};

	if((tm.tm_min = _sf_natoui(dts, 2)) == -1) {
		if(dts[-1] == ':') return 0;
		if(!_sf_isspace(*dts)) return 0;
		goto ex;
	};

	dts+=2;

	if(!*dts) goto zone;
	if(*dts == ':') dts++;
	else {
		if(*dts == '+' || *dts == '-') goto zone;
		if(*dts == 'Z') goto ex;
	};

	if((tm.tm_sec = _sf_natoui(dts, 2)) == -1) {
		if(dts[-1] == ':') return 0;
		if(!_sf_isspace(*dts)) return 0;
		goto ex;
	};

	dts+=2;

	if(!*dts) {
		/* Time in this form without the gmt offset is the LOCAL time */
		return mktime(&tm);
	};
	while(*dts == ' ') dts++;

zone:

	if(!*dts) {
		/* Time in this form without the gmt offset is the LOCAL time */
		return mktime(&tm);
	};

	if(*dts == 'Z') {
#ifdef	HAVE_TM_ZONE
	return mktime(&tm) + tm.tm_gmtoff;
#else	/* HAVE_TM_ZONE */
	return mktime(&tm) - timezone;
#endif	/* HAVE_TM_ZONE */
	};
	if(*dts != '+' && *dts != '-') {
		if(_sf_isspace(*dts)) goto ex;
		return 0;
	};

	tmp = _sf_recognize_tz(dts);

	goto ex;
};

f3:

if(form == 3) {

	if((tm.tm_mon = _sf_recognize_month(dts)) != -1) {
		form = 4;
		goto f4;
	}

	while(*dts && *dts != ' ')
		dts++;

	if(!*dts)
		return 0;

	dts++;

	tm.tm_mon = _sf_recognize_month(dts);

	if(tm.tm_mon < 0)
		return 0;

	while(*dts >= 'A' || *dts == ' ') dts++;

	if(*dts && dts[1] == ' ') {
		tm.tm_mday = _sf_natoui(dts++, 1);
	} else {
		tm.tm_mday = _sf_natoui(dts, 2);
		dts+=2;
	};

	if(tm.tm_mday < 1)
		return 0;

	while(*dts && *dts == ' ') dts++;


	if((tmp = _sf_recognize_time(dts)) == -1)
		return 0;

	dts += ((unsigned int)tmp) >> 24;
	tm.tm_hour = (tmp >> 16) & 0xff;
	tm.tm_min = (tmp >> 8) & 0xff;
	tm.tm_sec = tmp & 0xff;

	while(*dts == ' ') dts++;
	if(!*dts) return 0;

	if((tm.tm_year = _sf_natoui(dts, 4)) < 0) {
		if((tm.tm_year = _sf_natoui(dts, 2)) < 0)
			return 0;
		if(tm.tm_year < 38)
			tm.tm_year += 100;
		dts+=2;
	} else {
		tm.tm_year -= 1900;
		if(tm.tm_year < 0)
			return 0;
		dts+=4;
	};

	while(*dts && *dts == ' ')
		dts++;

	tmp = _sf_recognize_tz(dts);

	goto ex;

}

f4:

if(form == 4) {
	/* We can assume that tm.tm_mon contains valid month */

	{
		int z = tm.tm_mon;
		time(&tmp);
		localtime_r(&tmp, &tm);
		tm.tm_mon = z;
		tm.tm_isdst=-1;
	}

	while(*dts && *dts != ' ')	/* Skip month */
		dts++;

	while(*dts && *dts == ' ')	/* Skip space before the date */
		dts++;

	if(!*dts)
		return 0;

	if(
		((tm.tm_mday = _sf_natoui(dts, 2)) == -1)
		&& ((tm.tm_mday = _sf_natoui(dts, 1)) == -1)
	)
		return 0;

	while(*dts && *dts != ' ')	/* Skip date */
		dts++;

	while(*dts && *dts == ' ')	/* Skip space before the time */
		dts++;

	if(!*dts)
		return 0;

	if((tmp = _sf_recognize_time(dts)) == -1)
		return 0;

	/*
	**	We don't care about dts now
	dts += ((unsigned int)tmp) >> 24;
	*/
	tm.tm_hour = (tmp >> 16) & 0xff;
	tm.tm_min = (tmp >> 8) & 0xff;
	tm.tm_sec = tmp & 0xff;

	tmp = -1;
	goto ex;
}

ex:

	if(tmp == -1) {
		/* Time in RFC-822 form without the gmt offset is the LOCAL time */
		return mktime(&tm);
	};

#ifdef	HAVE_TM_ZONE
	return mktime(&tm) + tm.tm_gmtoff - tmp;
#else	/* HAVE_TM_ZONE */
	return mktime(&tm) - timezone - tmp;
#endif	/* HAVE_TM_ZONE */

};





/******************/
/* TIME TO STRING */
/******************/

char *
timetostr(time_t value, int flags) {
#define	SF_INT_TS_SIZE	47
	static char *ts=NULL;
	static struct tm *tm=NULL;
    static char *wd[]={"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
    static char *mn[]={"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
	int uf=0;
	int ogmt=0;


	/* All-or-nothing */
	if(ts == NULL) {
		ts = (char *)malloc(SF_INT_TS_SIZE + 1);
		tm = (struct tm *)malloc(sizeof(struct tm));
		if(!ts || !tm) {
			fprintf(stderr, "STRFUNC: Memory allocation error in timetostr()\n");
			abort();
		};
		ts[SF_INT_TS_SIZE]='\0';
	};

	/* Splitting the value */
	if((flags & TFMT_LOCAL) && !(flags & TFMT_OLD822GMT))
		(void)localtime_r(&value, tm);
	else {
		(void)gmtime_r(&value, tm);
	};

	if(flags & TFMT_UF)
		uf=1;
	if(flags & TFMT_OLD822GMT)
		ogmt=1;

	flags = flags & (int)0xff;

	if(flags == TFMT_RFC822) {
		char *fmts;

#ifdef	HAVE_TM_ZONE
		if(!tm->tm_zone)
			uf = 0;
#else
#ifndef	HAVE_TZNAME
			uf = 0;
#endif
#endif

		if(ogmt)
			fmts = "%s, %02d %s %4d %02d:%02d:%02d GMT";
		else
			fmts = "%s, %02d %s %4d %02d:%02d:%02d";

		fmts = ts + snprintf(ts, SF_INT_TS_SIZE, fmts,
		wd[tm->tm_wday], tm->tm_mday, mn[tm->tm_mon],
		tm->tm_year + 1900, tm->tm_hour, tm->tm_min, tm->tm_sec);

		if(!ogmt) {
			fmts += snprintf(fmts, SF_INT_TS_SIZE - (fmts - ts),
				" %c%02ld%02ld",

#ifdef	HAVE_TM_ZONE
				(tm->tm_gmtoff<0)?'-':'+',
				(long)( (labs(tm->tm_gmtoff) % 86400)  / 3600 ),
				(long)( (labs(tm->tm_gmtoff) % 3600) / 60 ));
			if(uf)
			fmts += snprintf(fmts, SF_INT_TS_SIZE - (fmts - ts),
				" (%s)", tm->tm_zone);
#else
#ifdef	HAVE_TZNAME
				(timezone>0)?'-':'+',
				(long)( (labs(timezone) % 86400)  / 3600 ),
				(long)( (labs(timezone) % 3600) / 60 ));
			if(uf)
			fmts += snprintf(fmts, SF_INT_TS_SIZE - (fmts - ts),
				" (%s)", tzname[0]);
#else
				0, 0);
#endif
#endif /* HAVE_TM_ZONE */

		};

	} else if(flags == TFMT_ISO8601 || flags == TFMT_X208) {
		char *fmts;

		if(flags == TFMT_ISO8601) {
			if(uf)
				fmts="%04d-%02d-%02dT%02d:%02d:%02d";
			else
				fmts="%04d%02d%02dT%02d%02d%02d";
		} else {
			fmts="%04d%02d%02d%02d%02d%02d";
		}

		fmts = ts + snprintf(ts, SF_INT_TS_SIZE, fmts,
				tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
				tm->tm_hour, tm->tm_min, tm->tm_sec);

#ifdef	HAVE_TM_ZONE
		if(tm->tm_gmtoff)
			snprintf(fmts, SF_INT_TS_SIZE - (fmts - ts), "%c%02ld%02ld",
				(tm->tm_gmtoff<0)?'-':'+',
				(long)( (labs(tm->tm_gmtoff) % 86400)  / 3600 ),
				(long)( (labs(tm->tm_gmtoff) % 3600) / 60 ));
		else
			snprintf(fmts, SF_INT_TS_SIZE - (fmts - ts), "Z");
#else
		if(timezone)
			snprintf(fmts, SF_INT_TS_SIZE - (fmts - ts), "%c%02ld%02ld",
				(timezone>0)?'-':'+',
				(long)( (labs(timezone) % 86400)  / 3600 ),
				(long)( (labs(timezone) % 3600) / 60 ));
		else
			snprintf(fmts, SF_INT_TS_SIZE - (fmts - ts), "Z");
#endif /* HAVE_TM_ZONE */

	} else if(flags == TFMT_CTIME) {
		char *fmts;
		fmts = ts + snprintf(ts, SF_INT_TS_SIZE,
			"%s %s %02d %02d:%02d:%02d %04d",
			wd[tm->tm_wday], mn[tm->tm_mon], tm->tm_mday,
			tm->tm_hour, tm->tm_min, tm->tm_sec, tm->tm_year + 1900);

#ifdef	HAVE_TM_ZONE
		if(uf)
			snprintf(fmts, SF_INT_TS_SIZE - (fmts - ts), " %c%02ld%02ld",
				(tm->tm_gmtoff<0)?'-':'+',
				(long)( (tm->tm_gmtoff % 86400)  / 3600 ),
				(long)( (tm->tm_gmtoff % 3600) / 60 ));
#else
		if(uf)
			snprintf(fmts, SF_INT_TS_SIZE - (fmts - ts), " %c%02ld%02ld",
				(timezone>0)?'-':'+',
				(long)( (timezone % 86400)  / 3600 ),
				(long)( (timezone % 3600) / 60 ));
#endif /* HAVE_TM_ZONE */
	} else {
		sprintf(ts, "%ld", value);
	};

	return ts;
};

