/*-
 * Copyright (c) 2000, 2001 Lev Walkin <vlm@spelio.net.ru>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: format.c,v 1.7 2001/05/15 15:19:23 vlm Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "strfunc.h"

#define	CHECK	{ register size_t n=0; \
	while((p-buf)+len>=buflen) buflen=buflen<<(n=1); \
	if(n) { \
		n=(size_t)(p-buf); \
		p=(buf=(char *)sf_realloc(buf, buflen)) + n; \
	}; };

typedef struct {
	char *buf;
	size_t buflen;
	char *metas;
	char *endmetas;
	char **(**funcs)(char *, void *);
} fmt_base_internal;

static char *sf_nel[]={ "X", NULL };

char *
formatf(fmt_base *fb, char *fmt, void *key) {
	char *f;
	register char *buf, *p;
	register size_t buflen=0;
	char *mp;	/* MetaPointer */
	char *emp;	/* EndMetaPointer */
	char *eb;	/* Ending bracet */
	char *ex;	/* Extracted data */
	char *q, *c;	/* Question and colon in "<a>?<b>:<c>" case */
	int ab, qb, cb;	/* Is there any '$''s present in <a> or <b> or <c>? */
	int es; /* Is there any quoted characters? */
	char **fr;	/* Function result */
	char **(*func)(char *, void *);
	register int len;
	register fmt_base_internal *fbd=(fmt_base_internal *)fb;
	int num;
	register char *cw;	/* Concatenate with */
	register char *eqw;	/* Test equality with */
	int eoq;	/* Equality or not equality test */
	register char n;

	if(!fbd) { return ""; };

	if(!fmt) {
		if(fbd->buflen == 0) {
			char *p = sf_realloc(fbd->buf, 16);
			if(!p)
				return NULL;
			fbd->buflen = 16;
		}
		*(fbd->buf) = (char)0;
		return fbd->buf;
	};

	f=fmt=sf_strdup(fmt);

	if(!buflen)
		buf=(char *)sf_malloc(buflen = 128);

	*(p=buf)=0;

while((n=*f)) {

	len = 0;

	/* Found quoting */
	if(n == '\\') {
		CHECK;
		switch((n=*++f)) {
			case 'n': *p++='\n'; break;
			case 't': *p++='\t'; break;
			case 's': *p++=' '; break;
			default: *p++=n;
		};
		f++;
		continue;
	};

	/* Metacharacter isn't found */
	if(n != '$' || !fbd->metas) {
		CHECK;
		*p++=n;
		f++;
		continue;
	};

	n=*++f;

	for(mp=fbd->metas, emp=fbd->endmetas; *mp; mp++, emp++)
		if(*mp == n) break;

	if(!*mp) {
		len=2;
		CHECK;
		*p++='$';
		*p++=n;
		f++;
		continue;
	};

	func=fbd->funcs[mp - fbd->metas];

	if(!(eb=brace(f++, *emp))) break;

	*eb=0;
	ex=f;

	/*
	 *      ex cw eb
	 *      |   | |
	 * abc${def+, }ghi
	 */

	num = -1;

	/* Is it an "a?b:c" case? */
	ab=qb=cb=es=0;
	for(f=ex, q=c=NULL; *f; f++) {
		switch(*f) {
			case '\\':
				if(!q && !c) es=1;
				f++;
				break;
			case '?':
				if(q) q=c=NULL;
				else q=f;
				break;
			case ':':
				if(c) q=c=NULL;
				else c=f;
				break;
			case '[':
				if(num == -1) {
					num = atoi(f+1);
					if(num != 0 || *(f+1) == '0')
						*f=0;
					else
						num=-1;
				}
				break;
			case '$': {
				register char *t;
				if(!q && !c) ab=f-ex+1;
				else if(c) cb=f-ex+1;
				else if(q) qb=f-ex+1;

				f++;
				/* Is there any non-escaped metacharacters? */
				if(!(t=strchr(fbd->metas, *f))) break;
				t=fbd->endmetas + (t - fbd->metas);
				f=brace(f, *t);

				}
				break;
		}
	}


	if(!q || (c && q > c)) {
		cw=ex; while(*cw && *cw != '+') cw++;
		if(*cw == '+' && cw[-1] != '\\') *cw++=0;
		else cw=NULL;

		fr=func(((ab&qb&cb)|es)?formatf(fbd, ex, key):ex, key);
		if(!fr) goto end;

		if(num < 0) {
			len=strlen(f=join(fr, cw?cw:", "));
			CHECK;
			strcpy(p, f);
		} else {
			register char **vp=fr;
			while(*vp && num) { vp++; num--; };
			if(num) goto end;
			len=strlen(*vp);
			CHECK;
			strcpy(p, *vp);
		};
	} else {
		*q++='\0';
		if(c) *c++='\0';

		eoq=0;
		eqw=ex-1;
		while(*++eqw) {
			if(*eqw == '!' && eqw[1] == '=') {
				*eqw=0;
				eqw+=2;
				eoq=-1;
				break;
			};
			if(*eqw == '=' && eqw[1] == '=') {
				*eqw=0;
				eqw+=2;
				eoq=1;
				break;
			};
		};

		if(eoq) {
			if(ab && ex+ab < eqw)
				fr=func(formatf(fbd, ex, key), key);
			else if(ab) {
				eqw=formatf(fbd, eqw, key);
				fr=func(es?formatf(fbd, ex, key):ex, key);
			} else {
				fr=func(es?formatf(fbd, ex, key):ex, key);
			};
			if(eoq == 1) {	/* Test of equality */
				if(!fr) {
					if(!*eqw) fr=sf_nel;
					goto here;
				};
				/* Fr filled */
				if(cfind(fr, eqw) == -1)  {
					fr=NULL; goto here; };
			} else {	/* Test of unequality */
				if(!fr) {
					if(*ecq) fr=sf_nel;
					goto here;
				};
				/* Fr filled */
				if(cfind(fr, eqw) != -1)  {
					fr=NULL; goto here; };
			};
		} else {
			fr=func((ab|es)?formatf(fbd, ex, key):ex, key);
		};
here:
		if(!fr || !*fr) {
			if(c) {
				len=strlen(f=cb?formatf(fbd, c, key):c);
				CHECK;
				strcpy(p, f);
			};
		} else {
			len=strlen(f=qb?formatf(fbd, q, key):q);
			CHECK;
			strcpy(p, f);
		};

	};

	p+=len;

end:

	f=eb+1;
	CHECK;
};

	*p=0;

	free(fmt);

	fbd->buflen=p-buf;
	if(fbd->buf) free(fbd->buf);
	return fbd->buf=buf;
};

fmt_base *
format_init() {

	return (fmt_base *)sf_calloc(sizeof(fmt_base_internal), 1);

}

void
format_free(fmt_base *fb) {
	fmt_base_internal *fbd=(fmt_base_internal *)fb;

	if(!fbd) return;
	if(fbd->buf) {
		free(fbd->buf);
	}
	if(fbd->metas) {
		free(fbd->metas);
		free(fbd->endmetas);
		free(fbd->funcs);
	}

	free(fbd);
}

int
format_metarule(fmt_base *fb, char sw, char ew, char **(*func)(char *, void *)) {
	fmt_base_internal *fbd=(fmt_base_internal *)fb;
	int len=0;

	if(!fbd || !sw || !ew) return -1;
	
	if(fbd->metas) len = strlen(fbd->metas);
	len++;

	fbd->metas=(char *)sf_realloc(fbd->metas, sizeof(char) * (len+1));
	fbd->endmetas=(char *)sf_realloc(fbd->endmetas, sizeof(char) * len);

	fbd->funcs=(char **(**)(char *, void *))sf_realloc(fbd->funcs, sizeof(char **(*)(char *, void *)) * len);

	len--;
	fbd->metas[len]=sw;
	fbd->metas[len+1]='\0';
	fbd->endmetas[len]=ew;
	fbd->funcs[len]=func;

	return 0;
};

char *
format_lastresult(fmt_base *fb, size_t *optSize) {
	register fmt_base_internal *fbp = (fmt_base_internal *)fb;

	if(!fbp) {
		if(optSize)
			*optSize = 0;
		return "";
	}

	if(fbp->buf) {
		*optSize = fbp->buflen;
		return fbp->buf;
	}

	*optSize = 0;

	return "";
};

size_t
format_lastsize(fmt_base *fb) {
	register fmt_base_internal *fbp = (fmt_base_internal *)fb;

	if(fbp && fbp->buf)
		return fbp->buflen;

	return 0;
}

char *
format_detach(fmt_base *fb, size_t *optSize) {
	register fmt_base_internal *fbp = (fmt_base_internal *)fb;
	char *s;

	if(!fbp) {
		if(optSize)
			*optSize = 0;
		return sf_strdup("");
	}

	s=fbp->buf;

	if(!s) {
		s = sf_strdup("");
		if(optSize)
			*optSize = 0;
	} else {
		free(fbp->buf);
		if(optSize)
			*optSize = fbp->buflen;
	}

	fbp->buf = NULL;
	fbp->buflen = 0;

	return s;
};

