/*-
 * Copyright (c) 1999, 2000, 2001 Lev Walkin <vlm@spelio.net.ru>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: sed.c,v 1.8 2001/05/15 15:19:23 vlm Exp $
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/types.h>
#include <regex.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "strfunc.h"

#define	SF_OPT_GLOBAL	1
#define	SF_OPT_RESULTS	2
#define	SF_OPT_NOZERO	4

typedef struct {
	int tyo;	/* 1 - s///, 2 - y///, 3 - // */
	int rexopt;
	int opt;	/* 2 - remember results, 1 - global matching, 0 - off. */
	regex_t *re;
	int nmatch;
	regmatch_t *pmatch;
	slist *to;
	sbuf  *sb;
	char *sfr;
	size_t sfr_len;
	char *sto;
	size_t sto_len;
	char *(*ResHandler)(slist *);
	slist *Results;
} sed_i;

sed_t *
sed_compile(char *expr) {
	char *c;
	char dlm;
	int tyo=0;
	int rexopt=0;
	int opt=0;
	char *strt, *pstrt;
	char *endt, *pendt;
	char *end, *pend;
	slist *sl;
	sed_i *st;

	if(!expr)
		return NULL;

	/* Searching for the Type Of Operation */
	switch(*expr) {
		case 's':
			tyo = 1;
			expr++;
			break;
		case 'y':
			tyo = 2;
			expr++;
			break;
		default:
			tyo = 3;
			break;
	};

	dlm = *expr++;
	if(!dlm || !*expr)
		return NULL;

	pstrt=strt=sf_malloc(strlen(expr) + 1);
	if(!strt)
		return NULL;
	*pstrt = 0;

	for(c = expr; *c; c++) {
		if(*c == '\\') {
			if(*++c == dlm) {
				*pstrt++ = dlm;
			} else {
				*pstrt++ = '\\';
				*pstrt++ = *c;
			};
			continue;
		};
		if(*c == dlm)
			break;
		*pstrt++ = *c;
	};

	*pstrt = 0;

	if(!*c) {
		free(strt);
		return NULL;
	};

	pend=end=sf_malloc(strlen(c) + 1);
	if(!end) {
		free(strt);
		return NULL;
	};
	*pend=0;

	pendt=endt=sf_malloc(strlen(c) + 1);
	if(!endt) {
		free(strt);
		free(end);
		return NULL;
	};
	*pendt=0;

	sl = sinit();

	if(tyo == 3 && *c == dlm && !strchr(c+1, dlm))
		goto skipsecond;

	for(c++; *c; c++) {
		if(*c == '\\') {
			c++;
			if(*c >= '0' || *c <= '9' || (*c == '@')) {
				*pend = 0;
				sadd_attach(sl, end, strlen(end));
				sadd2(sl, c, 1);
				pend=end=sf_malloc(strlen(c) + 1);
				*pend=0;
				*pendt++ = *c;
				continue;
			};
			switch (*c) {
				case 'n':
					*pend++ = '\n';
					*pendt++ = '\n';
					break;
				case 't':
					*pend++ = '\t';
					*pendt++ = '\t';
					break;
			};
			*pend++ = *c;
			*pendt++ = *c;
			continue;
		};
		if(*c == dlm)
			break;
		if(*c == '&') {
			*pend = 0;
			sadd_attach(sl, end, strlen(end));
			sadd2(sl, "0", 1);
			pend=end=sf_malloc(strlen(c) + 1);
			*pend=0;
			*pendt++ = *c;
			continue;
		};

		*pend++ = *c;
		*pendt++ = *c;
	};

skipsecond:

	*pend = 0;
	*pendt = 0;

	sadd_attach(sl, end, strlen(end));

	if(!*c) {
		free(strt);
		free(endt);
		sfree(sl);
		return NULL;
	};

	if(tyo == 3) {
		rexopt |= REG_NOSUB;
	};

	if(sl->count > 1) {
		rexopt &= ~REG_NOSUB;
		rexopt |= REG_EXTENDED;
	};

	for(c++; *c; c++)
	switch(*c) {
		case 'e':
			rexopt |= REG_EXTENDED;
			break;
		case 'i':
			rexopt |= REG_ICASE;
			break;
		case 'g':
			opt |= SF_OPT_GLOBAL;	/* Global matching */
			break;
		case 'n':
			opt |= SF_OPT_NOZERO;	/* Ignore 0's match */
			break;
		case 'r':
			rexopt &= ~REG_NOSUB;
			opt |= SF_OPT_RESULTS;	/* Remember results */
			break;
		default:
			free(strt);
			free(endt);
			sfree(sl);
			return NULL;
	};

	st = sf_calloc(1, sizeof(*st));
	st->to = sl;
	st->tyo = tyo;
	st->opt = opt;

	st->sb = sbuf_init();

	st->re = NULL;
	st->sfr = strt;
	st->sfr_len = strlen(strt);
	st->sto = endt;
	st->sto_len = strlen(endt);
	st->rexopt = rexopt;

	st->ResHandler = NULL;
	st->Results = NULL;

	if(tyo == 1 || tyo == 3) {
		st->re = (regex_t *)sf_malloc(sizeof(regex_t));
		if(!st->re) {
			sed_free(st);
			return NULL;
		};

		if(regcomp(st->re, strt, rexopt)) {
			sed_free(st);
			return NULL;
		};

		/* re_nsub + one for the whole string */
		st->nmatch = st->re->re_nsub + 1;

		st->pmatch = (regmatch_t *)sf_malloc(st->nmatch * sizeof(regmatch_t));
		if(!st->pmatch) {
			sed_free(st);
			return NULL;
		};
	} else if(tyo == 2) {
		if(rexopt & REG_ICASE) {
			for(c = st->sfr; *c; c++)
				*c = tolower(*c);
		};
	};

	return st;
};



/*
** sed_exec(): Executes compiled expression against the specified string.
*/

char *
sed_exec(sed_t *sp, char *string) {
	register sed_i *se=(sed_i *)sp;
	int slen = 0;

	if(!string || !se)
		return NULL;

	slen = strlen(string);
	sbuf_zero(se->sb);

	if((se->tyo == 1) || (se->tyo == 3)) {
		size_t nm = se->nmatch;
		regmatch_t *pm = se->pmatch;
		slist *results = se->Results;
		int notstart = 0;
		int i;
		char *c;
		int addedsmth = 0;	/* Is anything matched (for ///) */

		pm[0].rm_so = 0;
		pm[0].rm_eo = slen;

		/* We'll save results (()'s) here. */
		if((se->opt & SF_OPT_RESULTS) && !results)
			results = se->Results = sinit();

		if(results)
			sclear(results);

here1:

		/* Try to evaluate expression against given string */
		if(regexec(se->re, string, nm, pm, notstart)) {
			sbuf_add(se->sb, string);
			if((se->tyo == 3) && !addedsmth)
				return NULL;
			return se->sb->buf;
		} else {
			addedsmth = 1;
		}

		/* Prepare results array */
		if(results) {
			if(nm == 1) {	/* If no braces detected in expression */
				if(pm[0].rm_so > -1)
					sadd2(results,
						string + pm[0].rm_so,
						pm[0].rm_eo - pm[0].rm_so
					);
			} else  {
				if((se->opt & SF_OPT_NOZERO))
					i = 1;
				else
					i = 0;
				for(; i < nm && pm[i].rm_so > -1; i++) {
					sadd2(results,
						string + pm[i].rm_so,
						pm[i].rm_eo - pm[i].rm_so
					);
				}
			}
		}

		/* Add characters before matched substring */
		if(pm[0].rm_so != -1)
			sbuf_add2(se->sb, string, pm[0].rm_so);
		sbuf_add(se->sb, se->to->list[0]);

		for(i = 1; i < se->to->count; i+=2) {
			int t;
			c = se->to->list[i];
			if((*c == '@') && se->ResHandler) {
				sbuf_add(se->sb, se->ResHandler(results));
				continue;
			};
			t = atoi(c);
			if(
				(t < 0)
				|| (t >= nm)
				|| (pm[t].rm_eo == -1)
			) continue;

			sbuf_add2(se->sb,
				string + pm[t].rm_so,
				pm[t].rm_eo - pm[t].rm_so
			);
			sbuf_add(se->sb, se->to->list[i+1]);
		};

		if(se->opt & SF_OPT_GLOBAL) {
			notstart |= REG_NOTBOL;
			string += pm[0].rm_eo;
			goto here1;
		};

		sbuf_add(se->sb, string + pm[0].rm_eo);

		/* Add the rest of the string */

		return se->sb->buf;

	} else if(se->tyo == 2) {

		char *s, *c, *b;
		char *sfr;
		char cs;
		int i;

		sbuf_extend(se->sb, slen + 1);
		b = se->sb->buf;
		*b = 0;

		sfr = se->sfr;

		if(se->rexopt & REG_ICASE) {

			for(s = string; (cs=*s); s++) {
				cs = tolower(cs);
				for(c = sfr; *c; c++)
					if(cs == *c)
						break;
				if(!*c) {
					*b++ = *s;
					continue;
				};
				i = c - sfr;
				if(i > se->sto_len)
					break;
				*b++ = se->sto[i];
			};

		} else {

			for(s = string; (cs=*s); s++) {
				for(c = sfr; *c; c++)
					if(cs == *c)
						break;
				if(!*c) {
					*b++ = *s;
					continue;
				};
				i = c - sfr;
				if(i > se->sto_len)
					break;
				*b++ = se->sto[i];
			};

		}	/* if(se->rexopt & REG_ICASE) */

		*b = 0;

		return se->sb->buf;

	};

	return NULL;
};


/*
** Destroy sed_t matching structure.
*/

void
sed_free(sed_t *sp) {
	register sed_i *se=(sed_i *)sp;

	if(!se)
		return;

	if(se->tyo == 1 && se->re)
		regfree(se->re);

	sfree(se->to);

	if(se->sb)
		sbuf_free(se->sb);

	if(se->sfr)
		free(se->sfr);

	if(se->sto)
		free(se->sto);

	if(se->Results)
		sfree(se->Results);

	if(se->pmatch)
		free(se->pmatch);

	free(se);
};

int
sed_handler(sed_t *sp, char *(*ResHandler)(slist *)) {
	register sed_i *se=(sed_i *)sp;
	if(!se) return -1;

	se->opt |= SF_OPT_RESULTS;
	se->ResHandler = ResHandler;

	return 0;
};

slist *
sed_results(sed_t *sp) {
	register sed_i *se=(sed_i *)sp;

	if(!se) return NULL;

	return se->Results;

};


