/*-
 * Copyright (c) 1999, 2000, 2001 Lev Walkin <vlm@spelio.net.ru>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: simplechar.c,v 1.5 2001/05/15 15:19:23 vlm Exp $
 */

#include <string.h>
#include <stdlib.h>
#include "strfunc.h"

int
countchar(const char *s, char c) {
	int cnt = 0;
	char *p = (char *)s;

	if(!s || !*s)
		return 0;

	while(*p)
		if(*p++==c)
			cnt++;

	return cnt;
};

int
countchar2(const char *s, const char *c) {
	int cnt = 0;
	char *p = (char *)s;

	if(!s || !*s || !c)
		return 0;

	while(*p)
		if(strchr(c, *p++))
			cnt++;

	return cnt;
}

int
countstr(const char *b, const char *l) {
	int n;
	int len;

	if(!b || !l) return 0;

	if(!*b)
		return 0;

	if(!(len = strlen(l)))
		return 0;

	if(len == 1)
		return countchar(b, *l);

	for(n = 0; *b; b++) {
		if(
			(*(short *)b == *(short *)l)
			&& !strncmp(b, l, len)
		) {
			b += len - 1;
			n++;
		}
	}

	return n;
}


char *
genstr(int i) {
	static char buf[32];

	snprintf(buf, sizeof(buf), "%d", i);

	return buf;
}


char **
charize(const char *what) {
	char **vp;

	vp=(char **)sf_malloc(sizeof(char *) << 1);
	if(!vp) {
		return NULL;
	}

	if(!(*vp = sf_strdup((char *)what?(char *)what:""))) {
		free_values(vp);
		return NULL;
	}

	vp[1]=NULL;  

	return vp;
}

size_t
count_values(char **vp) {
	int z=0;

	if(!vp) return 0;

	while(*vp++) z++;

	return z;
}
		
void
free_values(char **vp) {
	char **p;

	if(!(p=vp))
		return;

	for(; *p; p++)
		free(*p);

	free(vp);
};

int
copy_values(char **fp, char ***tp) {
	char **vp;

	if(!tp || !fp)
		return 0;

	vp=*tp=(char **)sf_calloc((count_values(fp) + 4), sizeof(char *));
	if(!vp)
		return 0;

	for(; *fp; vp++, fp++)
		if(!(*vp = sf_strdup(*fp))) {
			free_values(vp);
			*tp = NULL;
			return -1;
		}

    return 1;
};


char *
brace(char *s, char c) {
	char f;
	int level=0;

	if(!s || !(f=*s))
		return NULL;

	while(*(++s))
		if((*s == c) && !(level--))
			return s;
		else if(*s == f)
			level++;
		

	return NULL;
}


char *
bracen(char *s, char c, size_t len) {
	char f;
	int level=0;

	if(!s || !(f=*s))
		return NULL;

	while(len--) {
		++s;
		if((*s == c) && !(level--))
			return s;
		else if(*s == f)
			level++;
	}
		

	return NULL;
}

