/*-
 * Copyright (c) 1999, 2000, 2001 Lev Walkin <vlm@spelio.net.ru>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: split.c,v 1.8 2001/05/15 15:19:23 vlm Exp $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <errno.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifdef	HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif

#include "strfunc.h"

/****************/
/* SPLIT STRING */
/****************/

/* split string and add its contents to the slist structure */
/* Split: full version */
int
splitf(slist *s, char *msg, char *dlm, int strict) {
	char *p = msg;
	char *w = NULL;
	int tokens = 0;
	int dlen;
	char cp, cd;

	if(!s || !p)
		return 0;

	if(!dlm) {
		if(strict)
			dlm = ":";
		else
			dlm = " \t\n\r";
	}
	cd = *dlm;
	dlen = strlen(dlm);

	if(strict & 2) {

		for(; (cp = *p); p++) {
			if((cd == cp) && (strncmp(p, dlm, dlen) == 0)) {
				if(w) {
					sadd2(s, w, p-w);
					w = NULL;
					tokens++;
				} else if(strict & 1) {
					sadd2(s, "", 0);
					tokens++;
				}
				p += dlen - 1;
			} else {
				if(!w)
					w = p;
			}
		}

	} else {

		for(; (cp = *p); p++) {
			if((cd == cp) || memchr(dlm, cp, dlen)) {
				if(w) {
					sadd2(s, w, p-w);
					w = NULL;
					tokens++;
				} else if(strict & 1) {
					sadd2(s, "", 0);
					tokens++;
				}
			} else {
				if(!w)
					w = p;
			}
		}

	}

	if(w) {
		sadd2(s, w, p-w);
		tokens++;
	}

	return tokens;
}

slist *
split(char *msg, char *delim, int strict) {
	slist *sl;

	sl = sinit();
	if(!sl)
		return NULL;

	splitf(sl, msg, delim, strict);

	return sl;
}


#define CHECKRANGE if(w-buf >= buflen) { \
	register int z=w-buf; \
	if(!(buf = (char *)sf_realloc(buf, buflen = buflen << 1))) {	\
		while(tokens--) sdel(s, tokens);	\
		return -1;	\
	} \
	w = buf + z;	\
	}

/* split string and add its contents to the slist structure */
int
splitquotable(slist *s, const char *msg) {
	char *buf=NULL;
	size_t buflen = 0;
	char *p=(char *)msg;
	char *w;
	int tokens=0;
	register char cp;

	int insng = 0;	/* Inside the single quotes */
	int indbl = 0;	/* Inside the double quotes */

	if(!s || !p)
		return 0;

	buflen=16;
	buf=(char *)sf_malloc(buflen);

	w=buf;

	while(1) {
		switch((cp=*p)) {
			case NULL:
				if(w-buf) {
					sadd2(s, buf, w-buf + 1);
					s->list[s->count-1][w-buf] = 0;
					if(insng) s->list[s->count-1][w-buf + 1] = 1;
					else if(indbl) s->list[s->count-1][w-buf + 1] = 2;
					s->lens[s->count-1] = w-buf;

					tokens++;
					w=buf;
				}
				if(buf)
					free(buf);
				return tokens;
			case ' ':
			case '\t':
			case '\n':
				if(indbl || insng) {
					*w++=cp;
				} else if(w-buf) {
					sadd2(s, buf, w-buf + 1);
					s->list[s->count-1][w-buf] = 0;
					if(insng) s->list[s->count-1][w-buf + 1] = 1;
					else if(indbl) s->list[s->count-1][w-buf + 1] = 2;
					s->lens[s->count-1] = w-buf;
					tokens++;
					w=buf;
				}
				break;
			case '\'':
				if(indbl) {
					*w++=cp;
				} else {
					if(insng) {
						sadd2(s, buf, w-buf + 1);
						s->list[s->count-1][w-buf] = 0;
						if(insng) s->list[s->count-1][w-buf + 1] = 1;
						else if(indbl) s->list[s->count-1][w-buf + 1] = 2;
						s->lens[s->count-1] = w-buf;
						tokens++;
						w=buf;
						insng = 0;
					} else {
						insng = 1;
					}
				};
				break;
			case '"':
				if(insng) {
					*w++=cp;
				} else {
					if(indbl) {
						sadd2(s, buf, w-buf + 1);
						s->list[s->count-1][w-buf] = 0;
						if(insng) s->list[s->count-1][w-buf + 1] = 1;
						else if(indbl) s->list[s->count-1][w-buf + 1] = 2;
						s->lens[s->count-1] = w-buf;
						tokens++;
						w=buf;
						indbl = 0;
					} else {
						indbl = 1;
					}
				};
				break;
			case '\\':
				cp = *++p;
				if(insng) {
					if(cp == '\'') {
						*w++='\'';
						break;
					}
					*w++='\\';
					CHECKRANGE;
					*w++=cp;
					break;
				}
				if(indbl) {
					switch(cp) {
						case '"': *w++='\''; break;
						case '\'': *w++='\''; break;
						case '\n': break;
						default:
							*w++='\\';
							CHECKRANGE;
							*w++=cp;
					}
					break;
				}
				switch(cp) {
					case '\n': break;
					case 'a': *w++='\a'; break;
					case 'b': if(w-buf) w--; break;
					case 'e': *w++='\e'; break;
					case 'f': *w++='\f'; break;
					case 'r': *w++='\r'; break;
					case 'n': *w++='\n'; break;
					case 't': *w++='\t'; break;
					case 'v': *w++='\v'; break;
					case '\\': *w++='\\'; break;
					default: *w++=cp; break;
				}
				break;
			default:
				*w++=cp;
		} /* switch (*) */
		CHECKRANGE;
		p++;
	}

	if(buf)
		free(buf);
	return tokens;
}


int
split_network(char *ipm, unsigned int *ip, unsigned int *mask) {
	unsigned long t;
	char *msk = NULL;

	if(!ip || !mask)
		return -1;

	if(!ipm) {
		*ip = 0;
		*mask = 0;
		return -1;
	};

	t = strlen(ipm) + 1;
	if(t > 32)	/* Too long address/mask */
		return -1;

	msk = (char *)alloca(t);
	memcpy(msk, ipm, t);
	ipm = msk;

	if((msk=strchr(ipm, '/')))
		*msk++ = 0;

	if(!sf_iaton(ipm, ip)) {
		*ip = *mask = 0;
		return -1;
	};

	if(!msk) {
		*mask = 0xffffffff;
		return 0;
	};

	t = countchar(msk, '.');
	if(t && t != 3) {
		*ip = *mask = 0;
		return -1;
	};
	if(t && !sf_iaton(msk, mask)) {
		*ip = *mask = 0;
		return -1;
	};
	if(!t) {
		errno=0;
		t = strtoul(msk, NULL, 10);
		if((t > 32) || errno) return -1;
		if(!t) *mask = 0;
		else *mask = htonl(0xffffffff << (32 - t));
	};

	*ip = *ip & *mask;

	return 0;
}

